package board;

import java.util.Stack;

import engine.Constants;

public abstract class Board {
	
	/**
	 * Holds a rows X columns grid representing the game state 
	 */
	protected int[][] grid = new int[Constants.nRows][Constants.nCols];
	

	/**
	 * Holds a stack with the moves of the game. A move can be pushed into 
     * the stuck when played on the board, or can be popped
	 */
	protected Stack<Integer> moves = new Stack<Integer>();
	
	/**
	 * Three states of a cell: ME - my cell, OPP - owned by the opponent  and an empty cell 
	 */
	protected final static int ME = 0, OPP = 1, EMPTY = 2;
	
	public void mergeGrid(Board board) {
		this.grid = board.grid;
	}
	
	/**	
	 * Initializes the board
	 */
	public Board() {
		resetBoard();
	}

	/**	
	 * initializes the grid with EMPTY cells
	 */
	private void resetBoard() {
		for (int row = 0; row < Constants.nRows; row++) {
			for (int col = 0; col < Constants.nCols; col++) {
				grid[row][col] = EMPTY;
			}
		}
	}

	/**	
	 * 
	 * @return returns a vector of features representing the board
	 */
	abstract public double[] getFeatureVector();
	
	/**
	 * @return	-1 <==> no winner yet, the game is still on
	 * 			0 <==> board owner won
	 * 			1 <==> opponent won
	 * 			2 <==> board full, tie
	 */
	abstract public int winner();

	/**
	 * Alters the board configuration by playing a move
	 * NB: Assumes the move is legal
	 * 
	 * @param move	0-indexed column number
	 * @param isOpponent	true <==> if we are setting the opponents piece
	 */
	public void setMove(int move, boolean isOpponent) {
		grid[getTop(move)][move]=isOpponent?OPP:ME;
		moves.push(move);
	}
	
	/**
	 * 
	 * @param col
	 * @return the lowest EMPTY row number
	 */
	public int getTop(int col) {
		int r = 0;
		for (; r<Constants.nRows; ++r)
			if (grid[r][col]==EMPTY)
				break;
		return r;
	}
	
	/**
	 * 
	 * @param move
	 * @return a boolean value indicating if the move is legal
	 */
	public boolean isLegalMove(int move) {
		return getTop(move) < Constants.nRows;
	}

	/**
	 * 
	 * @param isOpponent
	 * @param col
	 * @param row
	 * @return if grid[row][col] is a threat for player, i.e. if there were a piece in that position would that constitute four
	 */
	public boolean isThreat(boolean isOpponent, int col, int row) {
		
		int player = isOpponent ? OPP:ME;
		
		// sometimes we will want to know if the cell on the row *below* the first row is a threat
		if (row<=-1||row>=Constants.nRows) return false;
		
		int i,ew,ns,nwse,nesw;
		ew=ns=nwse=nesw=0;

		for(i=1; col-i>=0 && grid[row][col-i]==player; ++i, ++ew);
		for(i=1; col+i<Constants.nCols && grid[row][col+i]==player; ++i, ++ew);
		if (ew>=3)
			return true;

		for(i=1; row-i>=0 && grid[row-i][col]==player; ++i, ++ns);
		for(i=1; row+i<Constants.nRows && grid[row+i][col]==player; ++i, ++ns);
		if (ns>=3)
			return true;

		for(i=1; col-i>=0 && row-i>=0 && grid[row-i][col-i]==player; ++i, ++nesw);
		for(i=1; col+i<Constants.nCols && row+i<Constants.nRows && grid[row+i][col+i]==player; ++i, ++nesw);
		if (nesw>=3)
			return true;

		for(i=1; col-i>=0 && row+i<Constants.nRows && grid[row+i][col-i]==player; ++i, ++nwse);
		for(i=1; col+i<Constants.nCols && row-i>=0 && grid[row-i][col+i]==player; ++i, ++nwse);
		if (nwse>=3)
			return true;

		return false;
	}
	
	/**
	 * Deletes the top move from the stack and return its value, -1 if there is nothing left, i.e. the board is empty
	 * @return the top move from the stack
	 */
	public int popMove() {
		if (moves.isEmpty()) 
			return -1;
		int lastMove=moves.pop();
		grid[getTop(lastMove)-1][lastMove]=EMPTY;
		return lastMove;
	}

	/**
	 * Clears the board (and all internal mechanisms) so we can start a new game
	 */
	public void clear() {
		resetBoard();
		moves.clear();
	}
	
	/**
	 * Counts the number of similar pieces (i.e. of the same ownership) touching a piece that *would* potentially be placed in the column passed
	 * E.g.
	 * 
	 * Assume o is ours, and x is the opponent's pieces.
	 * 
	 * | | | | | | | |
	 * | | | | | | |o|
	 * | |o|x|x|x|o|o|
	 * ---------------
	 * 
	 * countAdjacent(0,false) == 1
	 * countAdjacent(1,false) == 1
	 * countAdjacent(2,false) == 1
	 * countAdjacent(3,false) == 0
	 * countAdjacent(4,false) == 1
	 * countAdjacent(5,false) == 3
	 * countAdjacent(6,false) == 1
	 * countAdjacent(0,true) == 0
	 * countAdjacent(1,true) == 1
	 * countAdjacent(2,true) == 2
	 * countAdjacent(3,true) == 3
	 * countAdjacent(4,true) == 2
	 * countAdjacent(5,true) == 1
	 * countAdjacent(6,true) == 0
	 * 
	 * @param column
	 * @param isOpponent
	 * @return number of adjacent similar pieces 
	 */
	public int countAdjacent(int column, boolean isOpponent) {
		int count = 0,
			row = getTop(column),
			player = isOpponent ? OPP : ME;

		if (row>0 && grid[row-1][column]==player)
			count++;
		if (row+1<Constants.nRows && grid[row+1][column]==player)
			count++;
		if (column>0 && grid[row][column-1]==1)
			count++;
		if (column+1<Constants.nCols && grid[row][column+1]==player)
			count++;
		if (row+1<Constants.nRows && column+1<Constants.nCols && grid[row+1][column+1]==player)
			count++;
		if (row+1<Constants.nRows && column>0 && grid[row+1][column-1]==player)
			count++;
		if (row>0 && column>0 && grid[row-1][column-1]==player)
			count++;
		if (row>0 && column+1<Constants.nCols && grid[row-1][column+1]==player)
			count++;
		
		return count;
	}
}
