package board;

import engine.Constants;

public class BoardCount extends Board {

	public double[] getFeatureVector() {
		count();
		return new double[]{
			singles.getCount()[0][0],
			//singles.getCount()[0][1],
			doubles.getCount()[0][0],
			triples.getCount()[0][0],
			triples.getCount()[0][1],
			fours.getCount()[0][0],
			singles.getCount()[1][0],
			//singles.getCount()[1][1],
			doubles.getCount()[1][0],
			triples.getCount()[1][0],
			triples.getCount()[1][1],
			fours.getCount()[1][0]
		};
	}

	protected Counter singles = getSinglesCounter(),
		doubles = getDoublesCounter(),
		triples = getTriplesCounter(),
		fours = getFoursCounter();

	/**
	 * 
	 * @return counter of singles split by the game time 
	 */
	protected Counter getSinglesCounter() {
		return new Counter() {
			{ n = new double[2][2]; } // split by game time
			@Override
			public void count(int firstEmptyPosition, int xshift, int yshift, int startCol, int startRow) {
				if (players[ME]==1 && players[EMPTY]==3)
					if (moves.size() < Constants.singlesGameSizeCountCutoff)
						++n[0][0];
					else
						++n[0][1];
				else if (players[OPP]==1 && players[EMPTY]==3)
					if (moves.size() < Constants.singlesGameSizeCountCutoff)
						++n[1][0];
					else
						++n[1][1];
			}
		};
	}

	/**
	 * 
	 * @return counter of doubles
	 */
	protected Counter getDoublesCounter() {
		return new Counter() {
			{ n = new double[2][1]; }
			@Override
			public void count(int firstEmptyPosition, int xshift, int yshift, int startCol, int startRow) {
				if (players[ME]==2 && players[EMPTY]==2)
					++n[0][0];
				else if (players[OPP]==2 && players[EMPTY]==2)
					++n[1][0];
			}
		};
	}

	final protected Counter getDummyCounter() {
		return new Counter() {
			{ n = new double[2][1]; }
			@Override
			public void count(int firstEmptyPosition, int xshift, int yshift, int startCol, int startRow) {
			}
		};
	}
	
	/**
	 * 
	 * @return counter of triplets
	 */
	protected Counter getTriplesCounter() {
		return new Counter() {
			{ n = new double[2][2]; } // even and odd
			@Override
			public void count(int firstEmptyPosition, int xshift, int yshift, int startCol, int startRow) {
				int emptyRow = startRow+yshift*firstEmptyPosition,
					emptyCol = startCol+xshift*firstEmptyPosition;
				if (players[ME]==3 && players[EMPTY]==1 && !isThreat(true, emptyCol, emptyRow-1))
					++n[0][emptyRow%2];
				else if (players[OPP]==3 && players[EMPTY]==1 && !isThreat(false, emptyCol, emptyRow-1))
					++n[1][emptyRow%2];
			}
		};
	}

	/**
	 * 
	 * @return counter of fours
	 */
	protected Counter getFoursCounter() {
		return new Counter() {
			{ n = new double[2][1]; }
			@Override
			public void count(int firstEmptyPosition, int xshift, int yshift, int startCol, int startRow) {
				if (players[ME]==4)
					++n[0][0];
				else if (players[OPP]==4)
					++n[1][0];
			}
		};
	}
		
	protected abstract static class Counter {
		protected double[][] n;
		protected abstract void count(int firstEmptyPosition, int xshift, int yshift, int col, int row);
		protected double[][] getCount() { return n; }
		protected void reset() { for(int i=0;i<n[0].length;++i) n[0][i]=n[1][i]=0; }
	}
	
	/**
	 * Counts number of unblocked singles, doubles and triplets in a row, column, or diagonal
	 */
	protected void count() {
		singles.reset();
		doubles.reset();
		triples.reset();
		fours.reset();
		countInDirection(0, 1);
		countInDirection(1, 0);
		countInDirection(1, 1);
		countInDirection(-1, 1);
	}

	// ALWAYS IN EXISTENCE OPTIMISATION
	protected int[] players = new int[3];
	
	/**
	 * Counts number of unblocked singles, doubles and triplets in a given orientation determined by xshift and yshift
	 * @param xshift
	 * @param yshift
	 */
	private void countInDirection(int xshift, int yshift) {
		int colMin = xshift == -1 ? 3 : 0,
			colMax = Constants.nCols - (xshift == 1 ? 3 : 0),
			rowMax = Constants.nRows-3*yshift,
			firstEmptyPosition;
		for(int col=colMin; col<colMax; ++col) {
			for(int row=0; row<rowMax; ++row) {
				firstEmptyPosition = (grid[row][col] == EMPTY ? 0 : (grid[row+yshift][col+xshift] == EMPTY ? 1 : (grid[row+yshift*2][col+xshift*2] == EMPTY ? 2 : 3)));
				players[0]=players[1]=players[2]=0;
				/*for (int i=0;i<4; ++i) {
					++players[grid[row+yshift*i][col+xshift*i]];
				}*/
				// ROLL OUT OPTIMISATION saves around 200ms per 10000 self-play games
				++players[grid[row][col]];
				++players[grid[row+yshift][col+xshift]];
				++players[grid[row+yshift*2][col+xshift*2]];
				++players[grid[row+yshift*3][col+xshift*3]];
				singles.count(firstEmptyPosition, xshift, yshift, col, row);
				doubles.count(firstEmptyPosition, xshift, yshift, col, row);
				triples.count(firstEmptyPosition, xshift, yshift, col, row);
				fours.count(firstEmptyPosition, xshift, yshift, col, row);
			}
		}
	}

	public int winner() {
		
		int winner;
		if ((winner = countFourInDirection(0, 1)) != EMPTY || (winner = countFourInDirection(1, 0)) != EMPTY || (winner = countFourInDirection(1, 1)) != EMPTY || (winner = countFourInDirection(-1, 1)) != EMPTY) {
			return winner == ME ? 0 : 1;
		}
		
		if (moves.size() == Constants.nCols*Constants.nRows)
			return 2; //tie
		
		return -1;
	}
	
	/**
	 * skimmed down version of countInDirection
	 * @param xshift
	 * @param yshift
	 * @return the player with connected four or EMPTY else
	 */
	private int countFourInDirection(int xshift, int yshift) {
		int colMin = xshift == -1 ? 3 : 0,
			colMax = Constants.nCols - (xshift == 1 ? 3 : 0),
			rowMax = Constants.nRows-3*yshift, g1,g2,g3,g4;
		for(int col=colMin; col<colMax; ++col) {
			for(int row=0; row<rowMax; ++row) {
				g1 = grid[row][col];
				g2 = grid[row+yshift][col+xshift];
				g3 = grid[row+yshift*2][col+xshift*2];
				g4 = grid[row+yshift*3][col+xshift*3];
				if (g1==g2&&g2==g3&&g3==g4) {
					if (g1==ME)
						return ME;
					else if (g1==OPP)
						return OPP;
				}
			}
		}
		return EMPTY;
	}
	
}