package runtime;

import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import board.Board;
import board.BoardCount;
import engine.Constants;
import engine.Sarsa;

public class Manager {

	/**
	 * Holds pairs of the board's class name and it's description
	 */
	static Map<String, String> boards = new HashMap<String,String>();
	/***
	 * Holds pairs of the board's class name and a structure representing it's training data
	 */
	static Map<String, TrainingData[]> trainingData = new HashMap<String,TrainingData[]>();
	
	public static void main(String[] args) {

		// ADDING A NEW Board CLASS IS AS SIMPLE AS ADDING ONE LINE HERE
		
		boards.put("BoardCount",	"Separate Count");
		boards.put("BoardSimple",	"Simple Single Count");
		//boards.put("Board_Merav",	"1's (split by vertical position) + 2's + 3's + 4's");
		
		ManagerGUI.buildGUI();
		
	}

	/**
	 * 
	 * @param board - the board class name
	 * @return new instance of board
	 */
	public static Board getBoard(String board) {

		try {
			return (Board) Class.forName("board."+board).getConstructor().newInstance();
		} catch(Exception e) {
			e.printStackTrace();
		}

		return null;
		
	}
	
	/**
	 * Initializes two sarsa agents that will play against each other and runs the training session 
	 * @param trainingBoard - the board class name
	 * @param trust
	 */
	static void train(String trainingBoard, int trust) {
		
		Board board = new BoardCount();
		Sarsa[] learners = new Sarsa[]{ new Sarsa(trainingBoard, trust), new Sarsa(trainingBoard, trust) };

		// the initial value here (curr) sets the player who plays first - so we can change it to train different weights for different types of Sarsa players
		int curr = 0,
			winner = 0,
			t=0;
		
		long time = System.currentTimeMillis();
		
		while(!learners[0].hasConverged() && ManagerGUI.isTraining()) {
		
			learners[0].start(0==curr);
			learners[1].start(1==curr);
			
			do {
				int move = learners[curr].getMove();
				board.setMove(move, curr==0);
	
				curr = 1-curr;
				learners[curr].setOpponentsMove(move);
				
				winner = board.winner();
			} while(winner == -1);
	
			learners[0].finish(winner);
			learners[1].finish(winner!=2 ? 1-winner : winner);
			
			board.clear();
			
			++t;
	
			if(t%Constants.outputInterval == 0) {
				long newTime = System.currentTimeMillis();
				ManagerGUI.displayConvergenceValues(t, learners[0].getWeights(), newTime-time, true);
				ManagerGUI.displayConvergenceValues(t, learners[1].getWeights(), newTime-time, false);
				time = newTime;
			}
			if(ManagerGUI.isTrainingAll() && t%Constants.trainAllSaveInterval == 0) {
				TrainingData data = new TrainingData(learners);
				ManagerGUI.saveInterval(data, trainingBoard);
			}
			
		}
		
		if(!ManagerGUI.isTrainingAll()) {
			ManagerGUI.finishedTraining(learners);
		}
		
	}

	/**
	 * 
	 * @param board - the board class name
	 * @return the name of a file where weights will be saved 
	 */
	private static String getBoardIndexFileName(String board) {
		return "savedWeights/"+board+".txt";
	}

	/**
	 * Loads the training parameters of a given board class from a file
	 * @param board - the board class name
	 */
	static void loadTraining(String board) {
		try{
			
			List<TrainingData> data = new ArrayList<TrainingData>();
	
			FileInputStream fin;
			String path = getBoardIndexFileName(board);
			try {
				fin = new FileInputStream(path);
			} catch (FileNotFoundException e) {
				File f = new File(path);
				f.getParentFile().mkdirs();
				f.createNewFile();
				fin = new FileInputStream(path);
			}
			boolean read = false;
			ObjectInputStream ois = null;
			try {
				ois = new ObjectInputStream(fin);
				read = true;
				while (data.add((TrainingData)ois.readObject()));
			}catch(EOFException e){}
			if (read) {
				ois.close();
				fin.close();
			}

			trainingData.put(board, data.toArray(new TrainingData[data.size()]));
			
		}catch(Exception e){e.printStackTrace();}
		
	}
	
	/**
	 * Loads the training parameters of all boards
	 */
	static void loadTraining() {
		for(String board : boards.keySet())
			loadTraining(board);
	}

	/**
	 * 
	 * A structure holding the date, timestamp, number of iterations, 
	 * trust and weights of a training session
	 *
	 */
	static class TrainingData implements Serializable {
		private static final long serialVersionUID = 1L;
		Date date;
		int t;
		int trust;
		double[][] weights;
		public TrainingData(int t, double[][] weights, Date date, int trust) {
			this.t = t;
			this.weights = weights;
			this.date = date;
			this.trust = trust;
		}
		public TrainingData(Sarsa[] learners) {
			this.t = learners[0].getIterations();
			this.weights = new double[][]{learners[0].getWeights(), learners[1].getWeights()};
			this.date = new Date();
			this.trust = learners[0].getTrustFn();
		}
	}
	
	/**
	 * Saves training parameters to a file
	 * @param w1 - first set of weights
	 * @param w2 - second set of weights
	 * @param t - number of iterations
	 * @param board - class name
	 * @param trust
	 */
	synchronized protected static void save(double[] w1, double[] w2, int t, String board, int trust) {
		try {
			TrainingData learned = new TrainingData(t, new double[][]{w1,w2}, new Date(), trust);
			
			FileOutputStream fout = new FileOutputStream(getBoardIndexFileName(board),true);
			ObjectOutputStream oos = trainingData.get(board).length > 0 ? new AppendingObjectOutputStream(fout) : new ObjectOutputStream(fout);
			
			oos.writeObject(learned);
			
			oos.close();
			fout.close();
		}
		catch (Exception e) { e.printStackTrace(); }
		
		loadTraining(board);
	}
	
	/**
	 * http://stackoverflow.com/questions/1194656/appending-to-an-objectoutputstream/1195078#1195078
	 *
	 */
	static private class AppendingObjectOutputStream extends ObjectOutputStream {

		public AppendingObjectOutputStream(OutputStream out) throws IOException {
			super(out);
		}

		@Override
		protected void writeStreamHeader() throws IOException {
			reset();
		}

	}

	/**
	 * Runs n games (player 1 vs. player 2) and presents their winning/ties statistics
	 * @param player1
	 * @param player2
	 * @param weights1
	 * @param weights2
	 */
	public static void test(String player1, String player2, double[][] weights1, double[][] weights2) {

		IPlayer p1 = player1.equals("Random")? new PlayerRandom() : player1.equals("Manual")? new PlayerManual(): new PlayerSarsa(weights1[0], weights1[1], false, player1);
		IPlayer p2 = player2.equals("Random")? new PlayerRandom() : player2.equals("Manual")? new PlayerManual(): new PlayerSarsa(weights2[0], weights2[1], false, player2);

		int numGames = getNumGames(player1, player2);
		double[] stats = winPercentage(new IPlayer[]{p1,p2}, numGames );
		
		String description1 = player1.equals("Random")||player1.equals("Manual")?"":" ("+boards.get(player1)+")";
		String description2 = player2.equals("Random")||player2.equals("Manual")?"":" ("+boards.get(player2)+")";
		
		ManagerGUI.displayResults(player1+description1+" vs. "+player2+description2+"\n\n" + stats[0] +"% wins ("+player1+"), "+ stats[1] +"% ties.\n\nOut of "+(int)stats[2]+" games.");
	}
	
	/**
	 * @param player1
	 * @param player2
	 * @return number of games to be played: 1 in case of manual playing or n games else
	 */
	public static int getNumGames(String player1, String player2) {
		return player1.equals("Manual")||player2.equals("Manual")?1:Constants.numTestsForWinningStatistics;
	}
	
	/**
	 * Runs N games and calculates statistics
	 * @param players
	 * @param N - number of games to be played
	 * @return an array containing the winning and ties percentages
	 */
	private static double[] winPercentage(IPlayer[] players, int N) {

		Board board = new BoardCount();
		
		int start = 0,
			winner = 0,
			n = 0,
			wins = 0,
			ties = 0,
			curr;
		p:
		while(n < N && isActive(players)) {
			
			//if (n==N/2)
			//	start=1-start;
			
			curr = start;
			
			players[0].start(0==curr);
			players[1].start(1==curr);
			
			do {
				int move = players[curr].getMove();
				if (!isActive(players)) break p;
				board.setMove(move, curr!=0);
	
				curr = 1-curr;
				players[curr].setOpponentsMove(move);
				
				winner = board.winner();
			} while(winner == -1);
	
			players[0].finish(winner);
			players[1].finish(winner!=2 ? 1-winner : winner);
			
			n++;
			if (winner==0)
				wins++;
			if (winner==2)
				ties++;
			
			board.clear();
			
			ManagerGUI.incTest();
			
		}
		
		dispose(players);
		
		return new double[]{ wins/(double)n * 100, ties/(double)n * 100, n};
		
	}

	/**
	 * Disposes players structures 
	 * @param players
	 */
	private static void dispose(IPlayer[] players) {
		if (players[0] instanceof PlayerManual)
			((PlayerManual)players[0]).dispose();
		if (players[1] instanceof PlayerManual)
			((PlayerManual)players[1]).dispose();
	}

	/**
	 * 
	 * @param players
	 * @return a boolean value indicating if the manual player is still active 
	 */
	public static boolean isActive(IPlayer[] players) {
		return ManagerGUI.isActive() &&
				(!(players[0] instanceof PlayerManual) || ((PlayerManual)players[0]).isActive()) &&
				(!(players[1] instanceof PlayerManual) || ((PlayerManual)players[1]).isActive());
	}

	/**
	 * 
	 * @return a boolean value indicating if the gui is still active
	 */
	public static boolean isActive() {
		return ManagerGUI.isActive();
	}

	static Results[][] results;
	static ExecutorService thread;

	/**
	 *  Tests all pairs of all saved agents add show statistics
	 */
	public static void testAll() {
		
		int n = 0;
		final List<TestPlayer> players = new ArrayList<TestPlayer>();
		
		for (Entry<String, TrainingData[]> e : trainingData.entrySet()) {
			n += e.getValue().length;
			for(TrainingData data : e.getValue()) {
				TestPlayer p = new TestPlayer();
				p.weights = new double[][]{data.weights[0], data.weights[1]};
				p.board = e.getKey();
				p.name = ManagerGUI.getWeightsDescription(data);
				players.add(p);
			}
		}
		
		results = new Results[n][n];
		for(int i=0;i<n;++i)
			for(int j=0;j<n;++j)
				results[i][j] = new Results();
		
		ManagerGUI.setTestAllN(n);
		thread = Executors.newFixedThreadPool(6);
		
		for(int ii=0;ii<n;++ii) {
			for(int jj=0;jj<n;++jj) {
				final int i=ii,j=jj;
				thread.execute(new Runnable() {
					@Override
					public void run() {
						double[] res = winPercentage(new IPlayer[]{
								new PlayerSarsa(players.get(i).weights[0],players.get(i).weights[1],false,players.get(i).board),
								new PlayerSarsa(players.get(j).weights[0],players.get(j).weights[1],false,players.get(j).board)
								}, Constants.numTestsForWinningStatistics);
						results[i][j].wins = res[0];
						results[i][j].ties = res[1];
						results[i][j].losses = 100-res[0]-res[1];
						results[i][j].p1 = players.get(i);
						results[i][j].p2 = players.get(j);
					}
				});
			}
		}

	}
	
	/**
	 * A structure holding the statistics data
	 */
	static class Results {
		double wins, losses, ties;
		TestPlayer p1,p2;
	}
	
	/**
	 * A structure holding the players data
	 */
	static class TestPlayer {
		double[][] weights;
		String name, board;
	}

}
